class Content:
    def __init__(self, key: str, *, scope: str = None):
        self._key = key
        self._scope = scope

    @property
    def key(self) -> str:
        return self._key

    @property
    def scope(self) -> str:
        return self._scope



from typing import Dict

from ._http import HttpResponse


class JsonResponse(HttpResponse):
    def __init__(self, status_code: int, data: Dict):
        super().__init__(status_code, data, headers={"Content-Type": "application/json"})



from typing import Optional

from lambler.resource.function._metadata import LambdaFunctionMetadata
from lambler.resource.role import Role, get_role_from_arn


class LambdaManager:
    def __init__(self, client):
        self._client = client

    def is_lambda_function_exists(self, name: str) -> bool:
        try:
            self.get_lambda_function(name)
        except self._client.exceptions.ResourceNotFoundException:
            return False
        return True

    def create(self, name: str, zip_file_path: str, handler_name: str,
               role: Role) -> LambdaFunctionMetadata:
        zip_file = _read_zip_file_content(zip_file_path)

        self._create_lambda_function(name, role, handler_name, zip_file)
        function_url = self._create_lambda_function_public_url(name)

        return LambdaFunctionMetadata(name, function_url, role)

    def delete(self, name: str):
        self._client.delete_function_url_config(
            FunctionName=name,
        )
        self._client.delete_function(
            FunctionName=name,
        )

    def get_lambda_function(self, name: str) -> Optional[LambdaFunctionMetadata]:
        try:
            response = self._client.get_function(
                FunctionName=name,
            )
        except self._client.exceptions.ResourceNotFoundException:
            return None

        role = get_role_from_arn(arn=response["Configuration"]["Role"])

        response = self._client.get_function_url_config(FunctionName=name)
        url = response["FunctionUrl"]

        return LambdaFunctionMetadata(name, url, role)

    def _create_lambda_function_public_url(self, name):
        response = self._client.create_function_url_config(
            FunctionName=name,
            AuthType="NONE",
        )
        function_url = response['FunctionUrl']

        self._client.add_permission(
            Action="lambda:InvokeFunctionUrl",
            FunctionName=name,
            Principal="*",
            StatementId="FunctionURLAllowPublicAccess",
            FunctionUrlAuthType="NONE",
        )

        return function_url

    def _create_lambda_function(self, name, role, handler_name, zip_file):
        self._client.create_function(
            FunctionName=name,
            Runtime='python3.9',
            Role=role.arn,
            Handler=handler_name,
            Code=dict(ZipFile=zip_file),
        )


def _read_zip_file_content(zip_file_path: str):
    with open(zip_file_path, 'rb') as file:
        zip_file = file.read()
    return zip_file



import json
from dataclasses import dataclass
from typing import Dict, Any


@dataclass
class RoleMetadata:
    name: str
    arn: str


class RoleManager:
    def __init__(self, client):
        self._client = client

    def create(self, name: str, policy: Dict[str, Any]) -> RoleMetadata:
        response = self._client.create_role(
            RoleName=name,
            AssumeRolePolicyDocument=json.dumps(policy),
        )
        return RoleMetadata(name, response["Role"]["Arn"])

    def attach_policy(self, name: str, policy_arn: str):
        self._client.attach_role_policy(
            PolicyArn=policy_arn,
            RoleName=name,
        )

    def delete_cascade(self, name: str):
        response = self._client.list_attached_role_policies(
            RoleName=name,
        )
        arn_list = [policy["PolicyArn"] for policy in response["AttachedPolicies"]]

        for arn in arn_list:
            self._client.detach_role_policy(RoleName=name, PolicyArn=arn)

        self._client.delete_role(RoleName=name)

    @staticmethod
    def from_arn(arn: str) -> RoleMetadata:
        name = arn.split("/")[-1]
        return RoleMetadata(name, arn)

    def get_role(self, name: str) -> RoleMetadata:
        response = self._client.get_role(RoleName=name)
        return RoleMetadata(name, response["Role"]["Arn"])



from typing import Dict, Any, List, Union, Optional

from .base import Handler
from .content import ContentProvider, ContentProviderSpace

ContentProviderAllowedType = Union[ContentProvider, Dict[Optional[str], ContentProvider]]


class Lambler:
    def __init__(self):
        self._handlers: List[Handler] = []
        self._content_providers: Optional[ContentProviderSpace] = None

    def handle(self, *handlers: Handler):
        self._handlers = handlers

        self._ensure_handler_with_content_provider()

    def __call__(self, event: Dict, context: Any):
        return self._handlers[0].handle(event, context)

    def use_content(self, provider: ContentProviderAllowedType):
        self._content_providers = _make_content_provider_space(provider)

        self._ensure_handler_with_content_provider()

    def _ensure_handler_with_content_provider(self):
        if self._content_providers is None:
            return

        for handler in self._handlers:
            handler.set_content_provider_space(self._content_providers)


def _make_content_provider_space(provider: ContentProviderAllowedType) -> ContentProviderSpace:
    if isinstance(provider, dict):
        return ContentProviderSpace.from_dict(provider)
    elif isinstance(provider, ContentProvider):
        return ContentProviderSpace.from_provider(provider)
    raise NotImplementedError()



from typing import Dict

from ._content import Content
from ._provider import ContentProviderSpace


def build_marked_params(content_providers: ContentProviderSpace, marker: Content) -> Dict:
    return content_providers.get(marker.scope).load(marker.key)



import re
from typing import Dict, List


class EndpointPath:
    def __init__(self, pattern: re.Pattern, keys: List[str]):
        self._pattern = pattern
        self._keys = keys

    @classmethod
    def create(cls, path: str) -> 'EndpointPath':
        keys = re.findall(r"\{(.*?)}", path)
        pattern_str = '^' + re.sub(r"\{.*?}", r"(.*?)", path) + '/?$'
        pattern = re.compile(pattern_str)
        return cls(pattern, keys)

    def match(self, path: str) -> (int, bool):
        matched = self._pattern.match(path)
        if matched is None:
            return 0, False

        start, end = matched.span()
        length = end - start + 1
        return length, True

    def extract_params(self, path: str) -> Dict[str, str]:
        params = self._pattern.findall(path)
        return {key: params[i] for i, key in enumerate(self._keys)}



from ._http import HttpResponse


class HtmlResponse(HttpResponse):
    def __init__(self, status_code: int, data: str):
        super().__init__(status_code, data, headers={"Content-Type": "text/html; charset=UTF-8"})



import inspect
from typing import Any

from lambler.http._event import HttpEvent


class Header:
    def __init__(self, key: str):
        self._key = key

    def __repr__(self) -> str:
        return f"Header({self._key!r})"

    @staticmethod
    def validate_param(param: inspect.Parameter) -> None:
        type_ = param.annotation

        if not issubclass(type_, str):
            raise TypeError("Header marker can only be used with type 'str'")

    def extract_event(self, event: HttpEvent) -> Any:
        return event.headers[self._key]



import inspect
from typing import Callable, Any, Dict, Type, get_origin, get_args

from ._path import EndpointPath
from .._event import HttpEvent
from .._header import Header
from .._param import Param
from .._query import Query
from ... import content, template
from ...content import Content, ContentProviderSpace
from ...template import Template


class EndpointExecutor:
    def __init__(self, path: EndpointPath, path_length: int, f: Callable, signature: inspect.Signature,
                 event: HttpEvent, *,
                 content_providers: ContentProviderSpace = None):
        self.path_length = path_length
        self._path = path
        self._f = f
        self._signature = signature
        self._event = event

        self._content_providers = content_providers

    def execute(self) -> Any:
        return self._f(**self._extract_params())

    def _extract_params(self) -> Dict[str, Any]:
        if len(self._signature.parameters) == 0:
            return {}

        params = {name: self._extract_marker_value(marker=param.default, type_=param.annotation)
                  for name, param in self._signature.parameters.items()}

        return params

    def _extract_marker_value(self, marker: Any, type_: Type) -> Any:
        params = self._path.extract_params(self._event.path)

        if isinstance(marker, Header):
            value = marker.extract_event(self._event)
        elif isinstance(marker, Content):
            value = content.build_marked_params(self._content_providers, marker)
        elif isinstance(marker, Param):
            value = params[marker.key]
        elif isinstance(marker, Query):
            value = _extract_query(marker, self._event, type_)
        elif isinstance(marker, Template):
            value = template.build_marked_params(self._content_providers, type_)
        else:
            raise NotImplementedError()
        return value


def _extract_query(marker: Query, event: HttpEvent, type_: Type):
    value = event.query_params[marker.key]
    if type_ is list:
        return value.split(",")
    if get_origin(type_) is list:
        args = get_args(type_)
        if args is ():
            return value.split(",")
        assert len(args) == 1
        cast = args[0]
        return [cast(v) for v in value.split(",")]

    return type_(value)



from typing import Any, Dict


class HttpResponse:
    def __init__(self, status_code: int, body: Any, *, headers: Dict[str, str] = None):
        self._status_code = status_code
        self._body = body
        self._headers = headers or {}

    def to_dict(self) -> Dict[str, Any]:
        return {
            "statusCode": self._status_code,
            "headers": self._headers,
            "body": self._body,
            "cookies": [],
            "isBase64Encoded": False,
        }



import inspect
from typing import Callable, Dict, Any, Optional

from ._executor import EndpointExecutor
from ._path import EndpointPath
from .._event import HttpEvent
from .._header import Header
from ...content import ContentProviderSpace


class Endpoint:
    def __init__(self, path: EndpointPath, f: Callable, signature: inspect.Signature):
        self._path = path
        self._f = f
        self._signature = signature

        self._content_providers: Optional[ContentProviderSpace] = None

    @classmethod
    def create(cls, path: str, f: Callable) -> 'Endpoint':
        signature = inspect.signature(f)
        _validate_markers(signature)
        return cls(EndpointPath.create(path), f, signature)

    def match(self, event: Dict, _: Any) -> Optional[EndpointExecutor]:
        http_event = HttpEvent.from_dict(event)
        path_length, match = self._path.match(http_event.path)
        if not match:
            return None

        return EndpointExecutor(self._path, path_length, self._f, self._signature, http_event,
                                content_providers=self._content_providers)

    def set_content_provider_space(self, providers: ContentProviderSpace):
        self._content_providers = providers


def _validate_markers(signature: inspect.Signature):
    for name, param in signature.parameters.items():
        marker = param.default

        if isinstance(marker, Header):
            marker.validate_param(param)



import time

# noinspection PyPackageRequirements
import boto3

from lambler.resource.function._manager import LambdaManager
from lambler.resource.role import get_or_create_lambda_function_role
from ._function import LambdaFunction


def get_or_create_lambda_function(name: str, zip_file: str, handler_name: str) -> LambdaFunction:
    manager = LambdaManager(boto3.client("lambda"))
    fn = manager.get_lambda_function(name)
    if fn is None:
        role = get_or_create_lambda_function_role("LamblerBasicExecution")
        time.sleep(20)  # TODO: use wait-for or retry loop
        fn = manager.create(name, zip_file, handler_name, role)

    return LambdaFunction(manager, name, fn.function_url, fn.role)



from lambler.resource.function._manager import LambdaManager
from ..role import Role


class LambdaFunction:
    def __init__(self, manager: LambdaManager, name: str, url: str, role: Role):
        self._manager = manager
        self._name = name
        self._url = url
        self._role = role

    @property
    def url(self) -> str:
        return self._url

    def delete_cascade(self):
        self._manager.delete(self._name)
        self._role.delete()



from abc import ABC, abstractmethod
from typing import Any, Optional, Dict


class ContentProvider(ABC):

    @abstractmethod
    def load(self, key: str) -> Any:
        pass


class ContentProviderSpace:
    def __init__(self, mapper: Dict[Optional[str], ContentProvider]):
        self._mapper = mapper

    def get(self, scope: Optional[str]) -> ContentProvider:
        return self._mapper[scope]

    @classmethod
    def from_dict(cls, mapper: Dict[Optional[str], ContentProvider]) -> 'ContentProviderSpace':
        assert isinstance(mapper, dict)
        for key, value in mapper.items():
            assert (isinstance(key, str) or key is None)
            assert isinstance(value, ContentProvider)
        return cls(mapper)

    @classmethod
    def from_provider(cls, provider: ContentProvider) -> 'ContentProviderSpace':
        return cls.from_dict({None: provider})



from lambler.resource.role._manager import RoleManager


class Role:
    def __init__(self, manager: RoleManager, name: str, arn: str):
        self._manager = manager
        self._name = name
        self._arn = arn

    @property
    def arn(self) -> str:
        return self._arn

    def delete(self):
        self._manager.delete_cascade(self._name)

    def attach_policy(self, policy_arn: str):
        self._manager.attach_policy(self._name, policy_arn)



from typing import Dict, Any


class HttpEvent:
    def __init__(self, path: str, headers: Dict[str, Any], query_params: Dict[str, str]):
        self.path = path
        self.headers = headers
        self.query_params = query_params

    @classmethod
    def from_dict(cls, raw: Dict) -> 'HttpEvent':
        return cls(raw["rawPath"], raw["headers"], raw.get("queryStringParameters", {}))



from typing import Callable, Any, TypeVar, Dict, List

from lambler.base import Handler
from ._endpoint import Endpoint
from ._response import HttpResponse
from ..content import ContentProviderSpace

T = TypeVar("T", bound=Callable)


class HttpApi(Handler):
    def __init__(self):
        self._endpoints: List[Endpoint] = []

    def get(self, path: str) -> Callable[[Callable], Any]:
        def decorator(f: T) -> T:
            self._endpoints.append(Endpoint.create(path, f))
            return f

        return decorator

    def set_content_provider_space(self, providers: ContentProviderSpace):
        for endpoint in self._endpoints:
            endpoint.set_content_provider_space(providers)

    def handle(self, event: Dict, context: Any):
        longest_path_length = 0
        longest_path_executor = None

        for endpoint in self._endpoints:
            executor = endpoint.match(event, context)
            if executor is not None and executor.path_length > longest_path_length:
                longest_path_length = executor.path_length
                longest_path_executor = executor

        if longest_path_executor is not None:
            response = longest_path_executor.execute()
            if response is not None:
                assert isinstance(response, HttpResponse)
                return response.to_dict()
            return http_response(200, "")


def http_response(status_code: int, body: str) -> Dict:
    return {
        "statusCode": status_code,
        "headers": {
            "Content-Type": "text/plain",
        },
        "body": body,
        "cookies": [],
        "isBase64Encoded": False
    }



from abc import ABC, abstractmethod
from typing import Any, Dict

from lambler.content import ContentProviderSpace


class Handler(ABC):
    @abstractmethod
    def handle(self, event: Dict, context: Any):
        pass

    @abstractmethod
    def set_content_provider_space(self, providers: ContentProviderSpace):
        pass



# noinspection PyPackageRequirements
import boto3

from ._manager import RoleManager
from ._role import Role


def get_or_create_lambda_function_role(name: str) -> Role:
    client = boto3.client('iam')
    manager = RoleManager(client)
    try:
        metadata = manager.get_role(name)
        role = Role(manager, metadata.name, metadata.arn)
    except client.exceptions.NoSuchEntityException:
        role = create_lambda_function_role(name)

    return role


def create_lambda_function_role(name: str) -> Role:
    manager = RoleManager(boto3.client('iam'))

    role_policy = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "",
                "Effect": "Allow",
                "Principal": {
                    "Service": "lambda.amazonaws.com"
                },
                "Action": "sts:AssumeRole"
            }
        ]
    }

    metadata = manager.create(name, role_policy)

    role = Role(manager, metadata.name, metadata.arn)
    role.attach_policy("arn:aws:iam::aws:policy/service-role/AWSLambdaBasicExecutionRole")

    return role


def get_role_from_arn(arn: str) -> Role:
    manager = RoleManager(boto3.client('iam'))
    metadata = manager.from_arn(arn)
    return Role(manager, metadata.name, metadata.arn)



import inspect
from abc import abstractmethod, ABC
from typing import Dict, TypeVar, Type

from lambler import content
from lambler.content import ContentProviderSpace, Content
from ._template import Template

T = TypeVar("T", bound="TemplateBase")


def build_marked_params(content_providers: ContentProviderSpace, type_: Type):
    assert issubclass(type_, TemplateBase)
    return type_.do_load(content_providers)


class TemplateBase(ABC):

    @classmethod
    @abstractmethod
    def load(cls: Type[T], *args, **kwargs) -> T:
        pass

    @classmethod
    def do_load(cls: Type[T], content_providers: ContentProviderSpace) -> T:
        return cls.load(**_extract_template_params(inspect.signature(cls.load), content_providers))


def _extract_template_params(signature: inspect.Signature, content_providers: ContentProviderSpace) -> Dict:
    params = {}
    for name, param in signature.parameters.items():
        type_ = param.annotation
        marker = param.default
        if isinstance(marker, Content):
            value = content.build_marked_params(content_providers, marker)
        elif isinstance(marker, Template):
            value = build_marked_params(content_providers, type_)
        else:
            raise NotImplementedError()
        params[name] = value
    return params
